import pygame
import os
import sys

from random import choice
from settings import *
import os
import pygame

from math import sqrt

""" Dies ist die Basisklasse welche eine Bewegungsrichtung fuer einen Kreis beschreibt """
class Direction:
    """ Der Konstruktor """
    def __init__(self, imagesize):
        self.imagesize = imagesize
    
    """ Diese Funktion gibt das Bild der Position aus (benoetigt fuer die Anzeige des naechsten Kreises """
    def draw(self, dstImage):
        if self.image:
            dstImage.blit(self.image, ArrowPosition)

    """ Diese Funktion gibt die neue Position des kreises zurueck und wird von den eigentlichen Unterklassen ueberschrieben"""
    def move(self, oldpos):
        return oldpos

    """ Diese Funktion dient dazu eine Zufaellige Directionsklasse auszuwaehlen, @staticmethod bedeutet, dass hierfuer keine Klasseninstanz benoetigt wird"""
    @staticmethod
    def random(imagesize):
        direction = choice((DirectionUp, DirectionDown, DirectionLeft, DirectionRight))
        return direction(imagesize)

""" Fuer einen Kreis, welcher sich immer nach oben bewegt"""
class DirectionUp(Direction):
    image = pygame.image.load(os.path.join(os.path.dirname(__file__),  "data", "ArrowUp.png"))
    
    """ Anpassung der Kreisposition """
    def move(self, oldpos):
        y = oldpos[1]
        if y > 0:
            y -= 1
        else:
            y = Field[1] - self.imagesize.height
        return (oldpos[0], y)

""" Fuer einen Kreis, welcher sich immer nach Unten bewegt"""
class DirectionDown(Direction):
    image = pygame.image.load(os.path.join(os.path.dirname(__file__),  "data", "ArrowDown.png"))

    """ Anpassung der Kreisposition """
    def move(self, oldpos):
        y = oldpos[1]
        if y < Field[1] - self.imagesize.height:
            y += 1
        else:
            y = 0
        return (oldpos[0], y)
        
""" Fuer einen Kreis welcher sich immer nach links bewegt"""
class DirectionLeft(Direction):
    image = pygame.image.load(os.path.join(os.path.dirname(__file__),  "data", "ArrowLeft.png"))

    """ Anpassung der Kreisposition """
    def move(self, oldpos):
        x = oldpos[0]

        if x > 0:
            x -= 1
        else:
            x = Field[0] - self.imagesize.width
        return (x, oldpos[1])

""" Fuer einen Kreis welcher sich immer nach rechts bewegt"""
class DirectionRight(Direction):
    image = pygame.image.load(os.path.join(os.path.dirname(__file__),  "data", "ArrowRight.png"))
    
    """ Anpassung der Kreisposition """
    def move(self, oldpos):
        x = oldpos[0]

        if x < Field[0] - self.imagesize.width:
            x += 1
        else:
            x = 0
        return (x, oldpos[1])

""" 
Die Klasse Circle beschreibt einen Kreis
Und wie dieser sich verhaelt
"""
class Circle:
    image = pygame.image.load(os.path.join(os.path.dirname(__file__),  "data", "circle.png"))

    def __init__(self, pos=(0.0,0.0), direction=Direction.random(None)):
        #Der Kreis hat eine Position an der er sich Befindet
        self.pos = pos
	#und eine Richtung in die er sich bewegt
        self.direction = direction

    """ Die Methode draw dient dazu den Kreis auf einem als Parameter uebergebenen Surface zu malen"""
    def draw(self, dstImg):
        if self.image:
            imgPos = self.image.get_rect()
            dstImg.blit(self.image, (Offset[0] + self.pos[0], Offset[1] + self.pos[1]))

    """ Die methode dient dazu aus einem Kreis den Mittelpunkt 
    und den Radius zu bestimmen, dies ist fuer die Kollisionserkennung noetig"""
    def calculate_center_and_r(self):
        center_x = self.pos[0] + (self.image.get_rect().width / 2.0)
        center_y = self.pos[1] + (self.image.get_rect().width / 2.0)
        r = (self.image.get_rect().width + self.image.get_rect().width) / 4.0
        return (center_x, center_y, r)
    
    """ Die Methode Move bewegt den Kreis, die Richtung ist durch die Instanz der "Direction" Klasse bestimmt """
    def move(self):
        self.pos = self.direction.move(self.pos)

    """ Diese funktion gibt zurueck ob dieser Kreis mit dem ihm uebergebenen Kreis kollidiert """
    def colides(self, other):
        #Berechne die Mittelpunkte und die Radien der beiden Kreise
        center_x, center_y, r = self.calculate_center_and_r()
        other_x, other_y, other_r = other.calculate_center_and_r()
        distX = center_x - other_x
        distY = center_y - other_y

        #Berechne mit dem Satz des Pythagoras die Distanz der Kreise
        dist = sqrt((distX * distX) + (distY * distY))
        
        #Wenn die Distanz kleiner oder gleich der Entfernung der Kreise ist sind sie kollidiert
        return dist <= (r + other_r)

""" Die eigentliche Spielklasse """
class CircleFun:

    """ Der Konstruktor der Klasse, welcher ggf. auch das Fenster oeffnet """
    def __init__(self, init_screen=True):
        #setz die Punkte auf null
        self.points = 0
        #das Spiel laeuft
        self.running = True
        #initialisiere eine Clock fuer die Frameratenbegrenzung (und ggf. FPS Anzeige)
        self.clock = pygame.time.Clock()
        #lade die Spieldaten
        self.load_data()
        #erstelle eine Schriftart zum Anzeigen der Punkte
        self.font = pygame.font.Font(None, 17)
        #initialisiere ggf. das Fenster
        if init_screen:
            self.brk = False
            self.screen = pygame.display.set_mode(Size)
        #erstell ein array fuer die Kreise
        self.circles = []
        #setze die naechste Richtung fuer einen Kreis
        self.next_direction = Direction.random(Circle.image.get_rect())
    
    """ Eine Funktion zum Laden der Daten """
    def load_data(self):
        #hole das Hintergrundbild aus dem Data ordner mit dem namen main.png
        self.background = pygame.image.load(os.path.join("data", "main.png"))
    
    """ Die Funktion wenn ein neuer Kreis hinzugefuegt werden soll """
    def append_circle(self, pos):
        
        #hole die Position des Kreises (Mausklick Position - Versatz Position)
        xpos = pos[0] - Offset[0]
        ypos = pos[1] - Offset[1]
        
        #Erstelle einen Kreis mit den Positionen und der naechsten Richtung
        c = Circle((xpos, ypos), self.next_direction)
        
        #Die Position muss Positiv sein und im Feld liegen, ausserdem darf der Kreis mit keinem anderen kollidieren
        if xpos >= 0 and xpos < Field[0] - c.image.get_rect().width and \
                ypos >= 0 and ypos < Field[1] - c.image.get_rect().height \
                and (not any([x.colides(c) for x in self.circles])):
            #wenn ja fuege diesen Kreis dem array hinzu
            self.circles.append(c)
            #erhoehe den Punktestand
            self.points += 1
            #und setze die naechste Richtung
            self.next_direction = Direction.random(Circle.image.get_rect())

    """ Das Event callback, hier werden die Mausclicks abgefangen und auf das Beenden reagiert """
    def handle_events(self):
        for event in pygame.event.get():

            #wurde das fenster geschlossen?
            if event.type == pygame.QUIT:
                #dann beende das Programm apprupt
                sys.exit()

            elif event.type == pygame.KEYDOWN: 
                if event.key == pygame.K_ESCAPE:
                    #bei escape gedrueckt, beende es sanft
                    self.running = False

            elif event.type == pygame.MOUSEBUTTONDOWN and event.button == 1:
                #linke Maustaste gedrueckt, dann fuege einen Kreis mit der momentaten Position hinzu
                self.append_circle(event.pos)

    """ Die Funktion welche sich um die gesammte darstellung kuemmert """
    def draw(self):
        #male das Hintergrundbild
        self.screen.blit(self.background, self.background.get_rect())
        #male das "Naechste Richtung" Symbol
        self.next_direction.draw(self.screen)
        #Schreibe den Punktestand
        self.draw_points()
        #gehe jeden Kreis durch
        for circle in self.circles:
            #und lass diesen sich selber auf das Bild malen
            circle.draw(self.screen)
        #Schreibe die Bilddaten direkt in den Buffer
        pygame.display.flip()
        #Limitiere die FPS auf 50
        self.clock.tick(50)

    """ Die Move Funktion kuemmert sich um die Bewegungen im Spiel """
    def move(self):
        #durchlaufe jeden Kreis
        for circle in self.circles:
            #lass jeden Kreis sich selbst in seine Richtung bewegen
            circle.move()

    """ Ueberpruefe ob Kreise kollidiert sind """
    def check(self):
        #durchlaufe jeden Kreis
        for circle in self.circles:
            #durchlaufe jeden Kreis
            for otherCircle in self.circles:
                #gucke ob der kreis nicht der eigene Kreis ist,
                #und auch nicht mit anderen kollidiert
                if circle != otherCircle and circle.colides(otherCircle):
                    #ansonnsten gehe zur Funktion 'end' dieser Klasse
                    self.end()

    """ Kreise sind Kollidiert, was tun? """
    def end(self):
        #Man koennte einen Statusfenster aufmachen oder dergleichen


        #Oder einfach das Spiel neustarten:
        self.__init__(False)

        #Oder da dies Geschicklichkeitsspiel so doof ist einfach das Spiel beenden:
        #self.running = False

    """ Die Hauptschleife """
    def mainloop(self):
        #wird durchlaufen so lange das Spiel laeuft
        while self.running:
            #bearbeite die events (Maus / Tastatureingaben / Beenden)
            self.handle_events()
            #bewege die Kreise
            self.move()
            #pruefe auf Kollisionen
            self.check()
            #und dann Male alles
            self.draw()

    """ Eine kleine Funktion zum malen des Punktestands """
    def draw_points(self):
        #Mache den String 'x Punkte' zu einem Bild
        text = self.font.render('{0} Punkte'.format(self.points), True, (255, 255, 255))
        #hole das Rechteck mit den Daten der Position
        textRect = text.get_rect()
        #Zentriere es auf die werte aus base/settings.py PointsPosition
        textRect.centerx = PointsPosition[0]
        textRect.centery = PointsPosition[1]
        #und Male es
        self.screen.blit(text, textRect)


#ist dies die ausgefuerte datei?
if __name__ == "__main__":
    #dann initialisiere pygame
    pygame.init()
    #erstelle eine Klasse des Spiels
    circlefun = CircleFun()
    #und durchlaufe die Hauptschleife
    circlefun.mainloop()


